import json
import shlex
from nclu.tests import parser

from nclu.tests.base import BaseTest
from nclu.tests.utils import original_open
from network_docopt.tab import tab_words_expand


class TestFRRMatchTab(BaseTest):

    def setUp(self):
        from nclu.plugins import frr
        super(TestFRRMatchTab, self).setUp()
        self.patcher.expect_file('/etc/netd.conf', '''
enable_frr=True
color_diffs=False''')
        self.patcher.expect_file(frr.ETC_FRR_DAEMONS, '''
zebra=yes''')
        self.patcher.expect_file(frr.ETC_FRR_CONF, '''
service integrated-vtysh-config
router bgp
neighbor 2800:1:4::21
''')

    def test_match_bgppeer_matches_compressed_ipv6(self):
        import nclu.NetDaemon  # noqa prevent circular import on startup
        from nclu.plugins import frr

        self.assertTrue(
            frr.matchtab.match_bgppeer('2800:1:4::21', False, True))
        self.assertTrue(
            frr.matchtab.match_bgppeer('2800:1:4::21', False, True))
        self.assertTrue(
            frr.matchtab.match_bgppeer('2800:01:4::21', False, True))
        self.assertTrue(
            frr.matchtab.match_bgppeer('2800:01:4::21', False, True))

    def test_tab_complete_with_bgppeer_compressed_ipv6(self):
        from nclu.NetDaemon import NetDaemon
        daemon = NetDaemon()
        daemon.nclu_parser.re_init()
        tokens, options = tab_words_expand(
            ('<bgppeer>',), False, '2800:1:4::21',
            daemon.nclu_parser.tab_complete_hook, {
                '<text-peer-group>': '2800:1:4::21',
                '<bgppeer>': '2800:1:4::21',
                'bgp': 'bgp', 'add': 'add',
                'neighbor': 'neighbor', 'net': 'net'
            })
        self.assertEqual(tokens, ['2800:1:4::21'])
        self.assertIn('2800:1:4::21', options)

    def test_tab_complete_with_bgppeer_regular_ipv6(self):
        from nclu.NetDaemon import NetDaemon
        daemon = NetDaemon()
        daemon.nclu_parser.re_init()
        tokens, options = tab_words_expand(
            ('<bgppeer>',), False, '2800:01:4::21',
            daemon.nclu_parser.tab_complete_hook, {
                '<text-peer-group>': '2800:01:4::21',
                '<bgppeer>': '2800:01:4::21',
                'bgp': 'bgp', 'add': 'add',
                'neighbor': 'neighbor', 'net': 'net'
            })
        self.assertEqual(tokens, ['2800:01:4::21'])
        self.assertIn('2800:01:4::21', options)


class TestFRR(BaseTest):

    def test_map_valid_json_output(self):
        import nclu.NetDaemon  # noqa prevent circular import on startup
        from nclu.plugins.frr.showclear import map_json_output
        result = json.loads(map_json_output({
            "ipv4 unicast": """{
                "foo": "bar"
            }"""
        }))
        self.assertEquals(result['ipv4 unicast']['foo'], 'bar')

    def test_map_invalid_json_output(self):
        import nclu.NetDaemon  # noqa prevent circular import on startup
        from nclu.plugins.frr.showclear import map_json_output
        result = json.loads(map_json_output({
            "ipv4 unicast": "% VRF blue not active\n",
            "ipv6 unicast": "% VRF blue not active\n"
        }))
        self.assertTrue('VRF' in result['ipv4 unicast'])
        self.assertTrue('VRF' in result['ipv6 unicast'])

    def test_copy_changes_owner(self):
        from nclu.plugins import frr
        self.patcher.expect_file(frr.ETC_FRR_CONF_SCRATCHPAD, '')
        frr.copy_scratchpad_to_frr_conf()
        self.assertEqual(len(self.patcher.records['getpwnam']), 1)
        self.assertEqual(
            self.patcher.records['getpwnam'][0]['args'][0], 'frr')
        self.assertEqual(len(self.patcher.records['getgrnam']), 1)
        self.assertEqual(
            self.patcher.records['getgrnam'][0]['args'][0], 'frr')


def get_show_commands():
    from nclu.plugins.frr import frr_docstring
    result = []
    for line in frr_docstring.__doc__.splitlines():
        line = line.strip()
        if (not line.startswith('frr show') and
                not line.startswith('frr clear')):
            continue
        line = line[4:]
        result.append('net ' + line)

    for expanded, options, original in parser.get_possible_commands(result):
        yield original, ' '.join(expanded), options


class TestFRRShowCommands(BaseTest):
    dump_data = False  # to dump command data

    def test_commands(self):
        '''
        Use this to dump a text file of current commands and what they
        generate. This can then be used to regression test
        the architecture redesign
        '''
        from nclu.NetDaemon import NetDaemon
        from nclu.plugins import frr

        self.patcher.expect_file('/etc/netd.conf', '''
enable_frr=True
color_diffs=False''')
        self.patcher.expect_file(frr.ETC_FRR_DAEMONS, '''
zebra=yes''')
        self.patcher.expect_file(frr.ETC_FRR_CONF, '''
service integrated-vtysh-config
router bgp
neighbor 1.2.3.4
ip as-path access-list LARRY permit 1234
ip community-list standard MARGE permit 100:100
ip extcommunity-list standard MARGE permit 100:100
ip prefix-list BOB seq 5 permit any
ipv6 prefix-list JASON seq 5 permit any
route-map FOO permit 10
''')

        daemon = NetDaemon()
        self.patcher.expect_command(
            '/bin/systemctl is-active frr.service', 'yes')
        self.patcher.expect_command(
            '/usr/bin/vtysh*',
            "Unable to find default BGP instance")
        self.patcher.expect_command(
            '/sbin/ip*', '')
        self.patcher.command_runs = []
        with original_open('.commands-list.txt', 'w') as fi:
            for original, show_cmd, options in get_show_commands():
                daemon.nclu_parser.re_init()
                success, result = daemon.nclu_parser.run(shlex.split(show_cmd))
                self.assertTrue(success, result)
                frr.execute_command(daemon)
                if self.dump_data:
                    fi.write(json.dumps({  # pragma: no cover
                        'command': original,
                        'example': show_cmd,
                        'options': options,
                        'system_calls': self.patcher.command_runs
                    }) + '\n')
                self.patcher.command_runs = []

    def test_correct_show_vrf_command(self):
        import nclu.NetDaemon  # noqa prevent circular import on startup
        from nclu.plugins.frr.showclear import exec_vtysh_show_or_clear
        # check that show bgp [vrf <text>] vni [json] is working properly
        self.patcher.expect_command(
            '/usr/bin/vtysh -c show bgp vrf foobar vni ipv4 unicast json',
            "{}")
        self.patcher.expect_command(
            '/usr/bin/vtysh -c show bgp vrf foobar vni ipv6 unicast json',
            "{}")
        assert exec_vtysh_show_or_clear({
            '<text>': 'foobar',
            'json': True,
            'vrf': True,
            'bgp': True,
            'vni': True
        }, 'show bgp vrf foobar vni json')
        self.patcher.expect_command(
            '/usr/bin/vtysh -c show bgp vrf foobar vni ipv4 unicast',
            "{}")
        self.patcher.expect_command(
            '/usr/bin/vtysh -c show bgp vrf foobar vni ipv6 unicast',
            "{}")
        assert exec_vtysh_show_or_clear({
            '<text>': 'foobar',
            'json': True,
            'vrf': True,
            'bgp': True,
            'vni': True
        }, 'show bgp vrf foobar vni')

    def test_add_neighbor_correctly_correctly_parsed_bgp_compressed_ipv6(self):
        from nclu.NetDaemon import NetDaemon
        from nclu.plugins import frr

        self.patcher.expect_file('/etc/netd.conf', '''
enable_frr=True
color_diffs=False''')
        self.patcher.expect_file(frr.ETC_FRR_DAEMONS, '''
zebra=yes''')
        self.patcher.expect_file(frr.ETC_FRR_CONF, '''
service integrated-vtysh-config
''')
        self.patcher.expect_command(
            '/bin/systemctl is-active frr.service', 'yes')
        self.patcher.expect_command('/bin/systemctl is-active frr', 'yes')

        self.patcher.expect_command(
            '/usr/bin/vtysh*', '')

        daemon = NetDaemon()
        self.assertTrue(self.execute_command(
            daemon, 'net add bgp autonomous-system 40201')[0])
        self.assertTrue(self.execute_command(
            daemon,
            'net add bgp neighbor 3800:01:4::21 remote-as internal')[0])
        self.patcher.expect_command(
            '/usr/bin/vtysh -m -f {}'.format(frr.ETC_FRR_CONF_SCRATCHPAD),
            '''frr defaults datacenter
service integrated-vtysh-config
password cn321
line vty
 exec-timeout 0 0

end
router bgp 40201
 neighbor 3800:1:4::21 remote-as internal

end
''')
        self.assertTrue(self.execute_command(
            daemon,
            'net add bgp neighbor 3800:01:4::21 '
            'capability extended-nexthop')[0])
        self.assertIn(
            'neighbor 3800:1:4::21 remote-as internal',
            self.patcher.file_expectations[frr.ETC_FRR_CONF_SCRATCHPAD])
        self.assertIn(
            'neighbor 3800:1:4::21 capability extended-nexthop',
            self.patcher.file_expectations[frr.ETC_FRR_CONF_SCRATCHPAD])
        self.assertNotIn(
            'neighbor 3800:1:4::21 remote-as external',
            self.patcher.file_expectations[frr.ETC_FRR_CONF_SCRATCHPAD])
        self.assertNotIn(
            'neighbor 3800:01:4::21 remote-as external',
            self.patcher.file_expectations[frr.ETC_FRR_CONF_SCRATCHPAD])


class TestFRRConfigCommands(BaseTest):
    config = '''
service integrated-vtysh-config
bgp route-map delay-timer 1
router bgp 10
neighbor 1.2.3.4
ip as-path access-list LARRY permit 1234
ip community-list standard MARGE permit 100:100
ip extcommunity-list standard MARGE permit 100:100
ip prefix-list BOB seq 5 permit any
ipv6 prefix-list JASON seq 5 permit any
route-map FOO permit 10
'''

    def test_get_config_commands_bgp_order(self):
        # net add bgp autonomous-system should be before other bgp commands
        from nclu.NetDaemon import NetDaemon
        from nclu.plugins import frr
        from nclu.plugins.frr.showclear import show_config_commands

        self.patcher.expect_file(frr.ETC_FRR_CONF, self.config)
        self.patcher.expect_file('/etc/netd.conf', '''
enable_frr=True
color_diffs=False''')
        self.patcher.expect_command(
            '/usr/bin/vtysh -m -f {}'.format(frr.ETC_FRR_CONF),
            self.config)

        # forces setting up some network docopt magic
        # otherwise it can not correctly check the commands
        NetDaemon()

        supported, unsupported = show_config_commands(True, None)

        # finally, make sure the net add bgp is before other bgp commands
        self.assertTrue(
            (supported.index('net add bgp autonomous-system 10') <
                supported.index('net add bgp route-map delay-timer 1')))
